from datetime import datetime
from typing import NamedTuple

from pydantic import BaseModel, field_validator


def default_str(string: str) -> str:
    """Coerce the default string value to a blank string."""
    if string is not None:
        return string
    else:
        return ""


class FollowData(BaseModel):
    follower_count: int
    following_users_count: int
    dismissing_users_count: int


class Identifier(BaseModel):
    identifier_type: str
    identifier_value: str

    _default_str = field_validator("identifier_type", "identifier_value", mode="before")(default_str)


class Marquee(BaseModel):
    artist_name: str
    segment: str

    _default_str = field_validator("artist_name", "segment", mode="before")(default_str)


class SearchQueries(BaseModel):
    platform: str
    search_time: datetime | None
    search_query: str
    search_interaction_URIs: list[str] | None

    _default_str = field_validator("platform", "search_query", mode="before")(default_str)


class UserData(BaseModel):
    username: str | None
    email: str
    country: str
    created_from_facebook: bool
    facebook_UID: str | None
    birthdate: datetime
    gender: str
    postal_code: str | None
    mobile_number: str | None
    mobile_operator: str | None
    mobile_brand: str | None
    creation_time: datetime

    _default_str = field_validator("email", "country", "gender", mode="before")(default_str)


class Track(NamedTuple):
    artist: str
    album: str
    track: str
    uri: str


class Album(NamedTuple):
    artist: str
    album: str
    uri: str


class Artist(NamedTuple):
    name: str
    uri: str


class Library(BaseModel):
    tracks: list[Track]
    albums: list[Album]
    shows: list | None  # unknown type my data is blank
    episodes: list | None  # unknown type my data is blank
    banned_tracks: list[Track] | None  # unknown type my data is blank
    artists: list[Artist]
    banned_artists: list[Artist] | None
    other: list[str] | None


class StreamingHistory(BaseModel):
    ts: datetime
    username: str
    platform: str
    ms_played: int | None
    conn_country: str
    ip_addr_decrypted: str
    user_agent_decrypted: str
    master_metadata_track_name: str | None
    master_metadata_album_artist_name: str | None
    master_metadata_album_album_name: str | None
    spotify_track_uri: str | None
    episode_name: str | None
    episode_show_name: str | None
    spotify_episode_uri: str | None
    reason_start: str | None
    reason_end: str | None
    shuffle: bool | None
    skipped: int | None
    offline: bool | None
    offline_timestamp: datetime | None
    incognito_mode: bool | None

    _default_str = field_validator(
        "username",
        "platform",
        "conn_country",
        "ip_addr_decrypted",
        "user_agent_decrypted",
        "reason_start",
        "reason_end",
        mode="before",
    )(default_str)



#!/usr/bin/env python
import argparse

from dlt import pipeline as dlt_pipeline
from pypeline_functions.google_takeout.sources import google_takeout_seed_gcs, google_takeout_seed_local


def google_takeout_seed_to_warehouse(seed_source: str, seed_path: str, dataset_name: str, destination: str) -> None:
    """Run the Google Takeout data seed to Postgres pipeline."""
    pipeline = dlt_pipeline(
        pipeline_name="google_takeout_seed_postgres",
        dataset_name=dataset_name,
        destination=destination,
    )

    if seed_source == "local":
        data = google_takeout_seed_local(seed_path)
    elif seed_source == "gcs":
        data = google_takeout_seed_gcs(seed_path)

    info = pipeline.run(data)
    print(info)


def main() -> None:  # noqa: D103
    parser = argparse.ArgumentParser(
        description="Transfers data from the Google Takeout data seed file to the Data Warehouse"
    )
    parser.add_argument(
        "--seed_source", type=str, required=True, choices=["local", "gcs"], help="where the data seed file is located"
    )
    parser.add_argument("--seed_path", type=str, required=True, help="file path where the data seed is stored")
    parser.add_argument(
        "--dataset_name", type=str, required=True, help="name of the dataset where the data will be loaded"
    )
    parser.add_argument(
        "--destination", type=str, required=True, choices=["postgres", "bigquery"], help="where the data will be saved"
    )

    args = parser.parse_args()
    google_takeout_seed_to_warehouse(args.seed_source, args.seed_path, args.dataset_name, args.destination)


if __name__ == "__main__":
    main()



import os
import sys

# Add the src directory to sys.path
src_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "pypeline_functions"))
if src_dir not in sys.path:
    sys.path.insert(0, src_dir)

import logging

# Setup logging for the package
logging.getLogger(__name__).addHandler(logging.NullHandler())

try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"

from pypeline_functions import google_takeout, spotify, utils

__all__ = [
    "google_takeout",
    "spotify",
    "utils",
]



from datetime import datetime
from typing import NamedTuple

from pydantic import BaseModel, field_validator


def default_str(string: str) -> str:
    """Coerce the default string value to a blank string."""
    if string is not None:
        return string
    else:
        return ""


class CandidateLocation(BaseModel):
    lat: float
    lng: float
    address: str | None
    name: str | None
    place_id: str
    semantic_type: str | None
    location_confidence: float | None  # missing in older (around 2014/15) history


class PlaceVisit(BaseModel):
    lat: float | None
    lng: float | None
    center_lat: float | None
    center_lng: float | None
    address: str | None
    name: str | None
    location_confidence: float | None
    calibrated_probability: float | None
    place_id: str | None
    start_time: datetime | None
    end_time: datetime | None
    device_tag: int | None
    candidate_locations: list[CandidateLocation]
    place_confidence: str | None
    place_visit_type: str | None
    visit_confidence: float | None
    edit_confirmation_status: str | None
    place_visit_importance: str | None


class ChromeHistory(BaseModel):
    title: str
    page_transition: str
    url: str
    time_usec: datetime

    _default_str = field_validator("title", "page_transition", "url")(default_str)


class Subtitles(NamedTuple):
    name: str
    url: str | None


class Details(NamedTuple):
    name: str


class Activity(BaseModel):
    header: str
    title: str
    time: datetime
    description: str | None
    titleUrl: str | None
    subtitles: list[Subtitles] | None
    details: list[Details] | None
    products: list[str] | None
    activityControls: list[str] | None



from pypeline_functions.google_takeout.extract_google_takeout_seed import main
from pypeline_functions.google_takeout.google_takeout_seed_to_bigquery import google_takeout_seed_to_bigquery
from pypeline_functions.google_takeout.models import (
    Activity,
    CandidateLocation,
    ChromeHistory,
    Details,
    PlaceVisit,
    Subtitles,
)
from pypeline_functions.google_takeout.parsers import GoogleTakeoutParser

__all__ = [
    "Activity",
    "CandidateLocation",
    "ChromeHistory",
    "Details",
    "GoogleTakeoutParser",
    "PlaceVisit",
    "Subtitles",
    "google_takeout_seed_to_bigquery",
    "main",
]



import json
from collections.abc import Iterable, Sequence

import dlt
from dlt.sources import DltResource

from pypeline_functions.google_takeout.models import Activity, ChromeHistory, PlaceVisit
from pypeline_functions.google_takeout.parsers import GoogleTakeoutParser
from pypeline_functions.utils.storage import GoogleCloudStorage, LocalStorage


@dlt.source
def google_takeout_seed_gcs(bucket_name: str) -> Sequence[DltResource]:
    """
    Extract data from the Google Takeout seed located in Google Cloud Storage.

    Parameters
    ----------
    bucket_name : str
        The name of the bucket that the seed is located in.
    """
    DATA_PATH = "google/takeout/"  # noqa: N806
    DATETIME_FORMAT = "%Y%m%dT%H%M%SZ"  # noqa: N806
    gcs = GoogleCloudStorage()
    google_takeout_parser = GoogleTakeoutParser()

    @dlt.resource(
        name="chrome_history", write_disposition="merge", primary_key=("time_usec", "title"), columns=ChromeHistory
    )
    def chrome_history() -> Iterable[ChromeHistory]:
        """Extract the latest chrome history data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, DATA_PATH, "Chrome/History.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            for datum in data.get("Browser History", []):
                yield google_takeout_parser.chrome_history_parser(datum)

    @dlt.resource(name="activity", write_disposition="merge", primary_key=("header", "title", "time"), columns=Activity)
    def activity() -> Iterable[Activity]:
        """Extract the latest activity data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, DATA_PATH, "MyActivity.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            for datum in data:
                yield google_takeout_parser.activity_parser(datum)

    @dlt.resource(
        name="location", write_disposition="merge", primary_key=("lat", "lng", "start_time"), columns=PlaceVisit
    )
    def location() -> Iterable[PlaceVisit]:
        """Extract the latest location data."""
        latest_seeds = gcs.get_latest_seeds(
            bucket_name, DATA_PATH, "Location History (Timeline)/Records.json", DATETIME_FORMAT
        )
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            if "placeVisit" in data:
                for datum in data:
                    yield google_takeout_parser.location_parser(datum)

    return chrome_history, activity, location


def google_takeout_seed_local(seed_path: str) -> Sequence[DltResource]:
    """
    Extract data from the Google Takeout seed located locally.

    Parameters
    ----------
    seed_path : str
        The file path to the extracted Google takeout seed.
    """
    local_storage = LocalStorage()
    google_takeout_parser = GoogleTakeoutParser()

    @dlt.resource(
        name="chrome_history", write_disposition="merge", primary_key=("time_usec", "title"), columns=ChromeHistory
    )
    def chrome_history() -> Iterable[ChromeHistory]:
        """Extract the latest chrome history data."""
        latest_seeds = local_storage.get_latest_seed(seed_path, "takeout-*", "Chrome/History.json")
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                for datum in data.get("Browser History", []):
                    yield google_takeout_parser.chrome_history_parser(datum)

    @dlt.resource(name="activity", write_disposition="merge", primary_key=("header", "title", "time"), columns=Activity)
    def activity() -> Iterable[Activity]:
        """Extract the latest activity data."""
        latest_seeds = local_storage.get_latest_seed(seed_path, "takeout-*", "MyActivity.json")
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                for datum in data:
                    yield google_takeout_parser.activity_parser(datum)

    @dlt.resource(
        name="location", write_disposition="merge", primary_key=("lat", "lng", "start_time"), columns=PlaceVisit
    )
    def location() -> Iterable[PlaceVisit]:
        """Extract the latest location data."""
        latest_seeds = local_storage.get_latest_seed(seed_path, "takeout-*", "Semantic Location History\\/.*\\.json")
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                for datum in data.get("timelineObjects", []):
                    if "placeVisit" in datum:
                        yield google_takeout_parser.location_parser(datum)

    return chrome_history, activity, location



from pypeline_functions.spotify.extract_spotify_seed import main
from pypeline_functions.spotify.models import (
    Album,
    Artist,
    FollowData,
    Identifier,
    Library,
    Marquee,
    SearchQueries,
    StreamingHistory,
    Track,
    UserData,
)
from pypeline_functions.spotify.parsers import SpotifyParser
from pypeline_functions.spotify.sources import spotify_seed_gcs, spotify_seed_local
from pypeline_functions.spotify.spotify_seed_to_bigquery import spotify_seed_to_bigquery

__all__ = [
    "Album",
    "Artist",
    "FollowData",
    "Identifier",
    "Library",
    "Marquee",
    "SearchQueries",
    "SpotifyParser",
    "StreamingHistory",
    "Track",
    "UserData",
    "main",
    "spotify_seed_gcs",
    "spotify_seed_local",
    "spotify_seed_to_bigquery",
]



from collections.abc import Iterable, Sequence
from datetime import UTC, datetime
from time import mktime

import dlt
from dlt.sources import DltResource
from feedparser import parse
from pypeline_functions.rss.models import RSSFeed


@dlt.source
def rss_feed(feed_url: str) -> Sequence[DltResource]:
    """
    Extract data from an RSS feed.

    Parameters
    ----------
    feed_url : str
        The URL of the RSS feed.
    """

    @dlt.resource(name="rss_feed", write_disposition="merge", primary_key=("feed_url", "link"))
    def rss_feed() -> Iterable[RSSFeed]:
        feed = parse(feed_url)
        for entry in feed.entries:
            thumbnail = entry.get(
                "media_thumbnail", entry.get("media_content", [{"url": "", "width": "", "height": ""}])
            )[0]
            yield {
                "feed_url": feed_url,
                "title": entry.get("title"),
                "link": entry.get("link"),
                "author": entry.get("author"),
                "summary": entry.get("summary"),
                "thumbnail_url": thumbnail.get("url"),
                "thumbnail_width": thumbnail.get("width"),
                "thumbnail_height": thumbnail.get("height"),
                "published": datetime.fromtimestamp(mktime(entry.get("published_parsed")), tz=UTC),
            }

    return rss_feed



#!/usr/bin/env python

import argparse

from dlt import pipeline as dlt_pipeline
from pypeline_functions.rss.sources import rss_feed


def rss_feed_to_postgres(feed_url: str, dataset_name: str, destination: str) -> None:
    """Run the RSS feed to Postgres pipeline."""
    pipeline = dlt_pipeline(
        pipeline_name="rss_feed_postgres",
        dataset_name=dataset_name,
        destination=destination,
    )

    data = rss_feed(feed_url)

    info = pipeline.run(data)
    print(info)


def main() -> None:  # noqa: D103
    parser = argparse.ArgumentParser(description="Transfers data from an RSS feed to PostgreSQL using dlt")
    parser.add_argument("--feed_url", type=str, required=True, help="URL of the RSS feed")
    parser.add_argument(
        "--dataset_name", type=str, required=True, help="name of the dataset where the data will be loaded"
    )
    parser.add_argument(
        "--destination", type=str, required=True, choices=["postgres", "bigquery"], help="where the data will be saved"
    )

    args = parser.parse_args()

    rss_feed_to_postgres(args.feed_url, args.dataset_name)


if __name__ == "__main__":
    main()



from datetime import UTC, datetime


class SpotifyParser:
    """
    Namespace for all the JSON parser functions for the Spotify data.

    Methods
    -------
    follow_data_parser(dct:dict) -> dict
        Parse and format the follow data.
    identifier_parser(dct:dict) -> dict
        Parse and format the identifier data.
    marquee_parser(dct:dict) -> dict
        Parse and format a single entry of the Marquee data.
    search_query_parser(dct:dict) -> dict
        Parse and format a single entry of the Search Query data.
    user_data_parser(dct:dict) -> dict
        Parse and format the User data.
    _track_parser(dct:dict) -> dict
        Parse and format a single entry of track data.
    _album_parser(dct:dict) -> dict
        Parse and format a single entry of album data.
    _artist_parser(dct:dict) -> dict
        Parse and format a single entry of artist data.
    library_parser(dct:dict) -> dict
        Parse and format a single entry of Music Library data.
    streaming_history_parser(dct:dict) -> dict
        Parse and format a single entry of Streaming History data.
    """

    def follow_data_parser(self, dct: dict) -> dict:
        """
        Parse and format the follow data.

        Parameters
        ----------
        dct : dict
            A dictionary representing the follow data.
        """
        output = {}

        output["follower_count"] = int(dct["followerCount"])
        output["following_users_count"] = int(dct["followingUsersCount"])
        output["dismissing_users_count"] = int(dct["dismissingUsersCount"])

        return output

    def identifier_parser(self, dct: dict) -> dict:
        """
        Parse and format the identifier data.

        Parameters
        ----------
        dct : dict
            A dictionary representing the identifier data.
        """
        output = {}

        output["identifier_type"] = dct["identifierType"]
        output["identifier_value"] = dct["identifierValue"]

        return output

    def marquee_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of the Marquee data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of the Marquee data.
        """
        output = {}

        output["artist_name"] = dct["artistName"]
        output["segment"] = dct["segment"]

        return output

    def search_query_parser(self, dct: dict) -> dict:
        """
        Parse and format the Search Query data.

        Parameters
        ----------
        dct : dict
            A dictionary representing the Search Query data.
        """
        output = {}

        output["platform"] = dct["platform"]
        if dct["searchTime"] is not None:
            output["search_time"] = datetime.strptime(dct["searchTime"][:19], "%Y-%m-%dT%H:%M:%S")  # noqa: DTZ007
        else:
            output["search_time"] = None
        output["search_query"] = dct["searchQuery"]
        if len(dct["searchInteractionURIs"]) == 0:
            output["search_interaction_URIs"] = None
        else:
            output["search_interaction_URIs"] = dct.get("searchInteractionURIs", None)

        return output

    def user_data_parser(self, dct: dict) -> dict:
        """
        Parse and format the user data.

        Parameters
        ----------
        dct : dict
            A dictionary representing the user data.
        """
        output = {}

        output["username"] = dct.get("username", None)
        output["email"] = dct["email"]
        output["country"] = dct["country"]
        output["created_from_facebook"] = dct["createdFromFacebook"]
        output["facebook_UID"] = dct.get("facebookUid", None)
        output["birthdate"] = datetime.strptime(dct["birthdate"][:10], "%Y-%m-%d")  # noqa: DTZ007
        output["gender"] = dct["gender"]
        output["postal_code"] = dct.get("postalCode", None)
        output["mobile_number"] = dct.get("mobileNumber", None)
        output["mobile_operator"] = dct.get("mobileOperator", None)
        output["mobile_brand"] = dct.get("mobileBrand", None)
        output["creation_time"] = dct.get("creationTime", None)

        return output

    def _track_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of track data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of track data.
        """
        output = {}

        output["artist"] = dct.get("artist", "")
        output["album"] = dct.get("album", "")
        output["track"] = dct.get("track", "")
        output["uri"] = dct.get("uri", "")

        return output

    def _album_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of album data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of album data.
        """
        output = {}

        output["artist"] = dct.get("artist", "")
        output["album"] = dct.get("album", "")
        output["uri"] = dct.get("uri", "")

        return output

    def _artist_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of artist data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of artist data.
        """
        output = {}

        output["name"] = dct.get("name", "")
        output["uri"] = dct.get("uri", "")

        return output

    def library_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of Music Library data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a signle entry of Music Library data.
        """
        output = {}

        output["tracks"] = [self._track_parser(datum) for datum in dct.get("tracks", [])]
        output["albums"] = [self._album_parser(datum) for datum in dct.get("albums", [])]
        output["shows"] = dct.get("shows", None)
        output["episodes"] = dct.get("episodes", None)
        if (len(dct["bannedTracks"]) == 0) | (dct["bannedTracks"] is None):
            output["banned_tracks"] = None
        else:
            output["banned_tracks"] = [self._track_parser(datum) for datum in dct.get("bannedTracks", [])]
        output["artists"] = [self._artist_parser(datum) for datum in dct.get("artists", [])]
        if (len(dct["bannedArtists"]) == 0) | (dct["bannedArtists"] is None):
            output["banned_artists"] = None
        else:
            output["banned_artists"] = [self._artist_parser(datum) for datum in dct.get("bannedArtists", [])]
        output["other"] = dct.get("other", None)

        return output

    def streaming_history_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of Streaming History data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of Streaming History data.
        """
        output = {}

        output["ts"] = datetime.strptime(dct["ts"], "%Y-%m-%dT%H:%M:%SZ")  # noqa: DTZ007
        output["username"] = dct.get("username", "")
        output["platform"] = dct.get("platform", "")
        output["ms_played"] = dct.get("ms_played", None)
        output["conn_country"] = dct.get("conn_country", "")
        output["ip_addr_decrypted"] = dct.get("ip_addr_decrypted", "")
        output["user_agent_decrypted"] = dct.get("user_agent_decrypted", "")
        output["master_metadata_track_name"] = dct.get("master_metadata_track_name", None)
        output["master_metadata_album_artist_name"] = dct.get("master_metadata_album_artist_name", None)
        output["master_metadata_album_album_name"] = dct.get("master_metadata_album_album_name", None)
        output["spotify_track_uri"] = dct.get("spotify_track_uri", None)
        output["episode_name"] = dct.get("episode_name", None)
        output["episode_show_name"] = dct.get("episode_show_name", None)
        output["spotify_episode_uri"] = dct.get("spotify_episode_uri", None)
        output["reason_start"] = dct.get("reason_start", None)
        output["reason_end"] = dct.get("reason_end", None)
        output["shuffle"] = dct.get("shuffle", None)
        output["skipped"] = dct.get("skipped", None)
        output["offline"] = dct.get("offline", None)
        if (dct["offline_timestamp"] == 0) | (dct["offline_timestamp"] is None):
            output["offline_timestamp"] = None
        else:
            output["offline_timestamp"] = datetime.fromtimestamp(dct.get("offline_timestamp", 0) / 10**6, UTC)
        output["incognito_mode"] = dct.get("incognito_mode", None)

        return output



#!/usr/bin/env python

import argparse

from pypeline_functions.utils.storage import GoogleCloudStorage, LocalStorage


def extract_google_takeout_seed_gcs(landing_bucket_name: str, landing_prefix: str) -> None:
    """Run the extraction pipeline for the Google Takeout data seed to GCS."""
    gcs = GoogleCloudStorage()

    prefix_filter = "google/takeout"

    if landing_prefix == "":
        blob_paths = gcs.extract_zip_files("data-seeds", prefix_filter, landing_bucket_name)
    else:
        blob_paths = gcs.extract_zip_files("data-seeds", prefix_filter, landing_bucket_name, landing_prefix)

    print(blob_paths)


def extract_google_takeout_seed_local(file_path: str, destination_path: str) -> None:
    """Run the extraction pipeline for the Google Takeout data seed to local storage."""
    ls = LocalStorage()

    blob_paths = ls.extract_zip_files(file_path, destination_path)

    print(blob_paths)


def main() -> None:  # noqa: D103
    parser = argparse.ArgumentParser(description="Extracts the google takeout seed (.zip) files.")

    subparsers = parser.add_subparsers(dest="storage_type", required=True)

    # Subcommand for GCS extraction
    gcs_parser = subparsers.add_parser("gcs", help="Extract to Google Cloud Storage")
    gcs_parser.add_argument(
        "--landing_bucket_name", type=str, required=True, help="name of the bucket to store the extracted data seeds"
    )
    gcs_parser.add_argument(
        "--landing_prefix",
        nargs="?",
        type=str,
        default="",
        help="prefix path location where the extract will be stored. \
            if undeclared it will use the same prefix path as the source",
    )

    # Subcommand for Local extraction
    local_parser = subparsers.add_parser("local", help="Extract to local storage")
    local_parser.add_argument("--file_path", type=str, required=True, help="Path to the the .zip file to extract")
    local_parser.add_argument(
        "--destination_path",
        type=str,
        required=True,
        help="Directory path where the extracted files will be stored locally",
    )

    args = parser.parse_args()

    if args.storage_type == "gcs":
        extract_google_takeout_seed_gcs(args.landing_bucket_name, args.landing_prefix)
    elif args.storage_type == "local":
        extract_google_takeout_seed_local(args.file_path, args.destination_path)


if __name__ == "__main__":
    main()



from datetime import datetime

from pydantic import BaseModel, field_validator


def default_str(string: str) -> str:
    """Coerce the default string value to a blank string."""
    if string is not None:
        return string
    else:
        return ""


class RSSFeed(BaseModel):
    feed_url: str
    title: str
    link: str
    author: str
    summary: str
    thumbnail_url: str
    thumbnail_width: str
    thumbnail_height: str
    published: datetime

    _default_str = field_validator(
        "feed_url", "title", "link", "author", "summary", "thumbnail_url", "thumbnail_width", "thumbnail_height"
    )(default_str)



import json
from collections.abc import Iterable, Sequence

import dlt
from dlt.sources import DltResource

from pypeline_functions.spotify.models import (
    FollowData,
    Identifier,
    Library,
    Marquee,
    SearchQueries,
    StreamingHistory,
    UserData,
)
from pypeline_functions.spotify.parsers import SpotifyParser
from pypeline_functions.utils.storage import GoogleCloudStorage, LocalStorage


@dlt.source
def spotify_seed_gcs(bucket_name: str) -> Sequence[DltResource]:
    """
    Extract data from the Spotify seed located in Google Cloud Storage.

    Parameters
    ----------
    bucket_name : str
        The name of the bucket that the seed is located in.
    """
    ACCOUNT_DATA_PATH = "spotify/account_data/"  # noqa: N806
    STREAMING_HISTORY_PATH = "spotify/streaming_history"  # noqa: N806
    DATETIME_FORMAT = "%Y%m%dT%H%M%S"  # noqa: N806
    gcs = GoogleCloudStorage()
    spotify_parser = SpotifyParser()

    @dlt.resource(name="follow_data", write_disposition="replace", columns=FollowData)
    def follow_data() -> Iterable[FollowData]:
        """Extract the latest follow data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, ACCOUNT_DATA_PATH, "Follow.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            yield spotify_parser.follow_data_parser(data)

    @dlt.resource(name="identifier", write_disposition="replace", columns=Identifier)
    def identifier() -> Iterable[Identifier]:
        """Extract the latest identifier data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, ACCOUNT_DATA_PATH, "Identifiers.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            yield spotify_parser.identifier_parser(data)

    @dlt.resource(name="marquee", write_disposition="replace", columns=Marquee)
    def marquee() -> Iterable[Marquee]:
        """Extract the latest marquee data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, ACCOUNT_DATA_PATH, "Marquee.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            for datum in data:
                yield spotify_parser.marquee_parser(datum)

    @dlt.resource(
        name="search_queries",
        write_disposition="merge",
        primary_key=("search_query", "search_time"),
        columns=SearchQueries,
    )
    def search_query() -> Iterable[SearchQueries]:
        """Extract the latest search query data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, ACCOUNT_DATA_PATH, "SearchQueries.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            for datum in data:
                yield spotify_parser.search_query_parser(datum)

    @dlt.resource(name="user_data", write_disposition="replace", columns=UserData)
    def user_data() -> Iterable[UserData]:
        """Extract the latest user data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, ACCOUNT_DATA_PATH, "Userdata.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            yield spotify_parser.user_data_parser(data)

    @dlt.resource(name="library", write_disposition="replace", columns=Library)
    def library() -> Iterable[Library]:
        """Extract the latest library data."""
        latest_seeds = gcs.get_latest_seeds(bucket_name, ACCOUNT_DATA_PATH, "YourLibrary.json", DATETIME_FORMAT)
        for seed in latest_seeds:
            content = seed.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            yield spotify_parser.library_parser(data)

    @dlt.resource(name="audio_streaming_history", write_disposition="merge", primary_key="ts", columns=StreamingHistory)
    def audio_streaming_history() -> Iterable[StreamingHistory]:
        """Extract the latest audio streaming history data."""
        blobs = gcs.list_blobs_with_prefix(bucket_name=bucket_name, prefix=STREAMING_HISTORY_PATH)
        streaming_history_files = [blob for blob in blobs if blob.name.endswith(".json") and "Audio" in blob.name]
        for f in streaming_history_files:
            content = f.download_as_string().decode("utf-8", "replace")
            data = json.loads(content)
            for datum in data:
                yield spotify_parser.streaming_history_parser(datum)

    return follow_data, identifier, marquee, user_data, library, search_query, audio_streaming_history


@dlt.source
def spotify_seed_local(seed_path: str) -> Sequence[DltResource]:
    """
    Extract data from the Spotify seed located locally.

    Parameters
    ----------
    seed_path : str
        The file path to the extracted Spotify seed.
    """
    local_storage = LocalStorage()
    spotify_parser = SpotifyParser()
    DATETIME_FORMAT = "%Y%m%dT%H%M%S"  # noqa: N806

    @dlt.resource(name="follow_data", write_disposition="replace", columns=FollowData)
    def follow_data() -> Iterable[FollowData]:
        """Extract the latest follow data."""
        latest_seeds = local_storage.get_latest_seed(
            seed_path, "spotify-account_data-*", "Follow.json", datetime_format=DATETIME_FORMAT
        )
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                yield spotify_parser.follow_data_parser(data)

    @dlt.resource(name="identifier", write_disposition="replace", columns=Identifier)
    def identifier() -> Iterable[Identifier]:
        """Extract the latest identifier data."""
        latest_seeds = local_storage.get_latest_seed(
            seed_path, "spotify-account_data-*", "Identifiers.json", datetime_format=DATETIME_FORMAT
        )
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                yield spotify_parser.identifier_parser(data)

    @dlt.resource(name="marquee", write_disposition="replace", columns=Marquee)
    def marquee() -> Iterable[Marquee]:
        """Extract the latest marquee data."""
        latest_seeds = local_storage.get_latest_seed(
            seed_path, "spotify-account_data-*", "Marquee.json", datetime_format=DATETIME_FORMAT
        )
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                for datum in data:
                    yield spotify_parser.marquee_parser(datum)

    @dlt.resource(
        name="search_queries",
        write_disposition="merge",
        primary_key=("search_query", "search_time"),
        columns=SearchQueries,
    )
    def search_query() -> Iterable[SearchQueries]:
        """Extract the latest search query data."""
        latest_seeds = local_storage.get_latest_seed(
            seed_path, "spotify-account_data-*", "SearchQueries.json", datetime_format=DATETIME_FORMAT
        )
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                for datum in data:
                    yield spotify_parser.search_query_parser(datum)

    @dlt.resource(name="user_data", write_disposition="replace", columns=UserData)
    def user_data() -> Iterable[UserData]:
        """Extract the latest user data."""
        latest_seeds = local_storage.get_latest_seed(
            seed_path, "spotify-account_data-*", "Userdata.json", datetime_format=DATETIME_FORMAT
        )
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                yield spotify_parser.user_data_parser(data)

    @dlt.resource(name="library", write_disposition="replace", columns=Library)
    def library() -> Iterable[Identifier]:
        """Extract the latest library data."""
        latest_seeds = local_storage.get_latest_seed(
            seed_path, "spotify-account_data-*", "YourLibrary.json", datetime_format=DATETIME_FORMAT
        )
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                yield spotify_parser.library_parser(data)

    @dlt.resource(name="audio_streaming_history", write_disposition="merge", primary_key="ts", columns=StreamingHistory)
    def audio_streaming_history() -> Iterable[StreamingHistory]:
        """Extract the latest audio streaming history data."""
        latest_seeds = local_storage.get_latest_seed(
            seed_path, "spotify-streaming_history-*", ".json", datetime_format=DATETIME_FORMAT
        )
        for seed in latest_seeds:
            with open(seed, "r", encoding="utf-8") as file:  # noqa: UP015
                data = json.load(file)
                for datum in data:
                    yield spotify_parser.streaming_history_parser(datum)

    return follow_data, identifier, marquee, user_data, library, search_query, audio_streaming_history



"""Utility function for connecting to Google Cloud Storage (GCS) as a target destination.

This module provides:
- TargetGoogleCloudStorage: class containing helper functions for authenticating, uploading and managing files in GCS.
"""

import glob
import json
import logging
import os
import re
import zipfile
from collections.abc import Iterable, Iterator
from datetime import UTC, datetime
from io import BytesIO

from google.api_core.exceptions import RequestRangeNotSatisfiable
from google.cloud.storage import Blob, Client, transfer_manager
from stream_unzip import stream_unzip


# TODO: add logging to all functions
class GoogleCloudStorage:
    """A helper class for authenticating, uploading and managing files in Google Cloud Storage.

    Attributes
    ----------
    name : str
        The name of the target for logging purposes.
    service_account_file : str
        The path to the service account credentials file.

    Methods
    -------
    authenticate() -> bool
        Authenticate the Google Cloud Storage client with the provided service account file.
    upload(soure_file:str, bucket:str, file_path:str) -> bool
        Upload a single file to Google Cloud Storage at the specified destination.
    get_list_blob_files(bucket_name:str, blob_name: str) -> list
        Fetch all blob files within a GCS bucket.
    upload_local_directory_to_gcs(directory_path:str, bucket_name:str, blob_name:str) -> None
        Upload a local directory and its contents to GCS at the specified destination.
    delete_files_in_blob(bucket_name:str, blob_name:str) -> None
        Delete files inside the specified blob directory in the GCS bucket.
    """

    def __init__(self) -> None:
        self.client = Client()

    # ----------------------------------- Upload -----------------------------------
    # inspiration: https://cloud.google.com/storage/docs/uploading-objects
    def upload(self, source_file: str, bucket_name: str, file_path: str) -> bool:
        """Upload a single file to Google Cloud Storage at the specified destination.

        Parameters
        ----------
        source_file : str
            The file path of the file to be uploaded.
        bucket_name : str
            The name of the bucket that the file will be uploaded in.
        file_path : str
            The file path where the `source_file` will be stored in the specified bucket.
        """
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(file_path)
        blob.upload_from_filename(source_file)

    def upload_many_blobs_with_transfer_manager(
        self, bucket_name: str, filenames: list[str], source_directory: str = "", workers: int = 8
    ) -> None:
        """Upload every file in a list to a bucket, concurrently in a process pool."""
        bucket = self.client.bucket(bucket_name)

        results = transfer_manager.upload_many_from_filenames(
            bucket, filenames, source_directory=source_directory, max_workers=workers
        )

        for name, result in zip(filenames, results, strict=False):
            # The results list is either `None` or an exception for each filename in
            # the input list, in order.

            if isinstance(result, Exception):
                print(f"Failed to upload {name} due to exception: {result}")
            else:
                print(f"Uploaded {name} to {bucket.name}.")

    def upload_directory_with_transfer_manager(self, bucket_name: str, source_directory: str, workers: int = 8) -> None:
        """Upload every file in a directory, including all files in subdirectories."""
        from pathlib import Path

        bucket = self.client.bucket(bucket_name)

        # recursively get all files in `directory` as Path objects.
        directory_as_path_obj = Path(source_directory)
        paths = directory_as_path_obj.rglob("*")

        # filter so the list only includes files, not directories themselves.
        file_paths = [path for path in paths if path.is_file()]

        # these paths are relative to the current working directory. Next, make them relative to `directory`
        relative_paths = [path.relative_to(source_directory) for path in file_paths]

        # convert them all to strings.
        string_paths = [str(path) for path in relative_paths]

        # TODO: plan how to do logging
        # print("Found {} files.".format(len(string_paths)))

        # Start the upload.
        results = transfer_manager.upload_many_from_filenames(
            bucket, string_paths, source_directory=source_directory, max_workers=workers
        )

        for name, result in zip(string_paths, results, strict=False):
            # The results list is either `None` or an exception for each filename in the input list, in order.

            if isinstance(result, Exception):
                logging.error(f"Failed to upload {name} due to exception: {result}")
                raise ExceptionGroup("File upload failed", result)
            else:
                logging.info(f"Uploaded {name} to {bucket.name}.")

    # ----------------------------------- Delete -----------------------------------

    def detele_file(self, bucket_name: str, blob_name: str) -> None:
        """Delete a blob from the bucket.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket that the file will be uploaded in.
        blob_name : str
            The name of the blob to search.
        """
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(blob_name)

        blob.delete()
        logging.info(f"Deleted blob: {blob_name}")

    def delete_files_with_prefix(self, bucket_name: str, prefix: str | None = None) -> None:
        """Use a batch request to delete a list of objects with the given prefix in a bucket.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket that the file will be uploaded in.
        prefix : str, optional
            The path prefix of the object to be deleted.
            If left unspecified the entire bucket's content will be deleted.
        """
        blobs_to_delete = self.list_blobs_with_prefix(bucket_name, prefix)
        with self.client.batch():
            for blob in blobs_to_delete:
                blob.delete()
                logging.info(f"Deleted blob: {blob.name}")
            logging.info("Files inside the specified GCS blob directory have been deleted successfully.")
            return True

    # ----------------------------------- List -----------------------------------

    def list_blob_files(self, bucket_name: str) -> Iterator:
        """Fetch all blob files within a GCS bucket.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket whose content will be listed.
        """
        try:
            blobs = self.client.list_blobs(bucket_name)
            return blobs

        # TODO: Improve exception handling.
        except Exception as e:
            logging.error(f"An error occurred while listing blobs: {e}")

    def list_blobs_with_prefix(self, bucket_name: str, prefix: str) -> Iterator:
        """List all the blobs in the bucket that begin with the prefix.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket whose content will be listed.
        prefix_filter : str
            The path prefix to filter the content that will be listed.
        """
        # Note: Client.list_blobs requires at least package version 1.17.0.
        blobs = list(self.client.list_blobs(bucket_name, prefix=prefix))

        return blobs

    # inspiration: https://github.com/googleapis/google-cloud-python/issues/920#issuecomment-653823847
    def list_subfolders(self, bucket_name: str, prefix: str) -> Iterator:
        """List all the subfolders in the bucket that begin with the prefix.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket whose content will be listed.
        prefix_filter : str
            The path prefix to filter the content that will be listed.
        """
        iterator = self.client.list_blobs(bucket_name, prefix=prefix, delimiter="/")
        prefixes = set()
        for page in iterator.pages:
            prefixes.update(page.prefixes)
        return prefixes

    # ----------------------------------- Misc. -----------------------------------

    @staticmethod
    def _zipped_chunks(blob: Blob) -> Iterable[BytesIO]:
        chunk_size = 536870912  # 500 MB
        start = 0
        while True:
            end = start + chunk_size - 1
            try:
                chunk = blob.download_as_bytes(start=start, end=end)
            except RequestRangeNotSatisfiable:
                break
            yield bytes(chunk)
            start += chunk_size

    def extract_zip_files(
        self, bucket_name: str, prefix_filter: str, landing_bucket_name: str, landing_prefix: str | None = None
    ) -> list[str]:
        """Extract all the .zip files from a bucket that begin with the prefix and save them to another bucket.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket that the file will be uploaded in.
        prefix_filter : str
            The path prefix of the `.zip` files to be extracted.
        """
        bucket = self.client.get_bucket(bucket_name)

        landing_bucket = self.client.get_bucket(landing_bucket_name)

        blobs = self.list_blobs_with_prefix(bucket_name, prefix_filter)
        blob_paths = [blob.name for blob in blobs]

        # NOTE: we can't use batching because the payload must be less than 10MB (https://cloud.google.com/storage/docs/batch#overview)
        for blob_path in blob_paths:
            if ".zip" not in blob_path:
                continue
            blob = bucket.blob(blob_path)

            for file_name, file_size, unzipped_chunks in stream_unzip(self._zipped_chunks(blob), chunk_size=536870912):
                name = file_name.decode("utf-8")
                print(f"Processing file: {name}")
                for chunk in unzipped_chunks:
                    if landing_prefix:
                        blob_destination = f"{landing_prefix.removesuffix('/')}/{name}"
                        landing_blob = landing_bucket.blob(blob_destination)
                    else:
                        blob_destination = f"{blob_path.removesuffix('.zip')}/{name}"
                        landing_blob = landing_bucket.blob(blob_destination)
                    landing_blob.upload_from_file(file_obj=BytesIO(chunk), size=file_size, content_type="text/plain")

        return blob_paths  # list of zip files extracted

    def convert_json_to_jsonl(self, bucket_name: str, prefix_filter: str) -> None:
        """Convert .json files to .jsonl in specified bucket that begin with the prefix.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket that the `.json` file is located in.
        prefix_filter : str
            The path prefix of the `.json` file paths to be converted.
        """
        bucket = self.client.get_bucket(bucket_name)

        landing_blobs = self.list_blobs_with_prefix(bucket_name, prefix_filter)
        landing_blob_paths = [blob.name for blob in landing_blobs]
        json_blobs = [p for p in landing_blob_paths if re.search(r"\.json$", p) is not None]

        for json_blob in json_blobs:
            blob = bucket.blob(json_blob)
            blob_content = blob.download_as_string().decode("utf-8", "replace")
            data = json.loads(blob_content)
            jsonl_blob = bucket.blob(json_blob + "l")
            if isinstance(data, dict):
                jsonl_blob.upload_from_string(json.dumps(data))
            elif isinstance(data, list):
                content = ""
                for datum in data:
                    content = content + json.dumps(datum) + "\n"
                jsonl_blob.upload_from_string(content)

    def download_blob_as_string(self, bucket_name: str, blob_path: str) -> str:
        """
        Download blob file as string.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket that the blob is located in.
        blob_path : str
            The path to the blob that will be downloaded.
        """
        bucket = self.client.get_bucket(bucket_name)

        blob = bucket.blob(blob_path)
        blob_content = blob.download_as_string().decode("utf-8", "replace")
        return blob_content

    def get_latest_seeds(self, bucket_name: str, prefix: str, file_name: str, datetime_format: str) -> Blob:
        """
        Locate the latest data seeds for a given file name suffix based on the provided datetime_format.

        Parameters
        ----------
        bucket_name : str
            The name of the bucket that the data seed is located in.
        prefix : str
            The path prefix to filter the search results.
        file_name : str
            The file name of the data seed that corresponds to the path suffix that you wish to match.
        datetime_format : str
            The datetime string format to match in order to find the latest data seed entry.
        """
        subfolders = self.list_subfolders(bucket_name, prefix)
        timestamps = []

        for folder in subfolders:
            t = folder.removeprefix(prefix).removesuffix("/")
            timestamp = datetime.strptime(t, datetime_format)  # noqa: DTZ007
            timestamps.append(timestamp)

        latest_timestamp = max(timestamps)
        latest_seed = f"{prefix}{latest_timestamp.strftime(datetime_format)}/"

        blobs = self.list_blobs_with_prefix(bucket_name, latest_seed)
        output = []
        for blob in blobs:
            if file_name in blob.name:
                output.append(blob)
        return output


class LocalStorage:
    """A helper class for managing local files.

    Methods
    -------
    extract_zip_files(file_path: str, destination_path: str) -> list[str]
        Extracts a zip file and saves it in the given destination path.
        Returns a list of file paths for each extracted file.
    """

    @staticmethod
    def _extract_timestamp_from_filename(file_path: str, datetime_format: str, delimeter: str) -> datetime:
        file_metadata = file_path.split("/")[-1].split(delimeter)
        for metadata in file_metadata:
            print(metadata)
            try:
                return datetime.strptime(metadata, datetime_format).astimezone(UTC)
            except ValueError:
                continue

        return datetime.min.replace(tzinfo=UTC)

    @staticmethod
    def _get_files_with_prefix(folder_path: str, path_prefix: str) -> list[str]:
        matching_files = []
        pattern = re.compile(path_prefix)
        for root, _, files in os.walk(folder_path):
            for file in files:
                full_path = os.path.join(root, file)
                if pattern.search(full_path):
                    matching_files.append(full_path)

        return matching_files

    def get_latest_seed(
        self,
        base_dir: str,
        seed_prefix: str,
        file_filter: str = ".*",
        datetime_format: str = "%Y%m%dT%H%M%SZ",
        delimeter: str = "-",
    ) -> [str]:
        """"""
        folder_pattern = os.path.join(base_dir, seed_prefix)
        folders = glob.glob(folder_pattern)

        if not folders:
            print(f"No folders matching pattern ({seed_prefix}) found in {base_dir}.")
            return ""

        latest_folder = ""
        max_timestamp = datetime.min.replace(tzinfo=UTC)
        for folder in folders:
            cur = self._extract_timestamp_from_filename(folder, datetime_format, delimeter)
            if cur > max_timestamp:
                max_timestamp = cur
                latest_folder = folder
                print(latest_folder)

        matched_files = self._get_files_with_prefix(latest_folder, file_filter)
        print(f"Matched files: {matched_files}")
        return matched_files

    @staticmethod
    def extract_zip_files(file_path: str, destination_path: str) -> list[str]:
        """Extact the .zip file at the provided file path into the given destination path.
        """ """

        Parameters
        ----------
        file_path : str
            The file path to the .zip file
        destionation_path : str
            The location where the extracted files will be saved.
        """
        if not os.path.exists(destination_path):
            os.makedirs(destination_path)

        extracted_files = []  # List to store paths of extracted files

        with zipfile.ZipFile(file_path, "r") as zip_ref:
            # Extract files one by one to save memory
            for file_name in zip_ref.namelist():
                file_path_full = os.path.join(destination_path, file_name)

                # If it's a directory, create it
                if file_name.endswith("/"):
                    os.makedirs(file_path_full, exist_ok=True)
                else:
                    os.makedirs(os.path.dirname(file_path_full), exist_ok=True)

                    # Extract each file one by one to save memory
                    with open(file_path_full, "wb") as f:
                        chunk_size = 536870912  # 500 MB
                        with zip_ref.open(file_name) as zip_file:
                            while True:
                                chunk = zip_file.read(chunk_size)
                                if not chunk:
                                    break
                                f.write(chunk)

                    # Add the full path of the extracted file to the list
                    extracted_files.append(file_path_full)

            print(f"Extraction complete to {destination_path}")
            return extracted_files



from datetime import UTC, datetime


class GoogleTakeoutParser:
    """Namespace for all the JSON parser functions for the Google Takeout data.

    Methods
    -------
    chrome_history_parser(dct:dict) -> dict
        Parse and format a single entry of the Google Chrome History data.
    activity_parser(dct:dict) -> dict
        Parse and format a single entry of the Google Activity data.
    _candidate_location_parser(dct:dict) -> dict
        Parse and format a single entry of the candidate locations from the Semantic Location History data.
    location_parser(dct:dict) -> dict
        Parse and format a single entry of the Semantic Location History data.
    """

    def chrome_history_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of the Google Chrome History data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of the Google Chrome History data.
        """
        dct["title"] = dct.get("title", "")
        dct["page_transition"] = dct.get("page_transition", "")
        # TODO: add HTTP sanitation by converting to HTTPS
        dct["url"] = dct.get("url", "")
        dct["time_usec"] = datetime.fromtimestamp(dct.get("time_usec", 0) / 10**6, UTC)
        return dct

    def activity_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of the Google Activity data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of the Google Activity data.
        """
        datetime_format = "%Y-%m-%dT%H:%M:%S.%fZ"
        dct["header"] = dct.get("header")
        dct["title"] = dct.get("title")
        for datetime_format in ["%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ"]:
            try:
                dct["time"] = datetime.strptime(dct.get("title", "1970-01-01:00:00:00Z"), datetime_format)  # noqa: DTZ007
                break
            except ValueError:
                pass
        dct["description"] = dct.get("description")
        dct["titleUrl"] = dct.get("titleUrl")
        subtitles = dct.get("subtitles", [])
        _subtitles = []
        for subtitle in subtitles:
            _subtitles.append({"name": subtitle.get("name", ""), "url": subtitle.get("url", None)})
        dct["subtitles"] = _subtitles
        details = dct.get("details", [])
        _details = []
        for detail in details:
            _details.append({"name": detail.get("name", "")})
        dct["details"] = _details
        dct["products"] = dct.get("products")
        dct["activityControls"] = dct.get("activityControls")
        return dct

    def _candidate_location_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of the candidate locations from the Semantic Location History data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of the candidate locations from the Semantic Location History data.
        """
        output = {}

        output["lat"] = dct["centerLatE7"]
        output["lng"] = dct["centerLngE7"]
        output["place_id"] = dct["placeId"]
        output["semantic_type"] = dct.get("semanticType", None)
        output["address"] = dct.get("address", None)
        output["name"] = dct.get("name", None)
        output["location_confidence"] = dct.get("locationConfidence", None)

        return output

    def location_parser(self, dct: dict) -> dict:
        """
        Parse and format a single entry of the Semantic Location History data.

        Parameters
        ----------
        dct : dict
            A dictionary representing a single entry of the Semantic Location History data.
        """
        output = {}
        datetime_formats = ["%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ"]

        place_visit = dct["placeVisit"]
        location = place_visit["location"]
        duration = place_visit["duration"]
        output["lat"] = location.get("latitudeE7", None)
        output["lng"] = location.get("longitudeE7", None)
        output["place_id"] = location.get("placeId", None)
        output["location_confidence"] = location.get("locationConfidence", None)
        output["address"] = location.get("address", None)
        output["name"] = location.get("name", None)
        output["calibrated_probability"] = location.get("calibratedProbability", None)
        output["device_tag"] = location.get("sourceInfo", {"deviceTag": None}).get("deviceTag", None)
        if duration.get("startTimestamp", None) is None:
            output["start_time"] = None
        else:
            for datetime_format in datetime_formats:
                try:
                    output["start_time"] = datetime.strptime(duration.get("startTimestamp"), datetime_format)  # noqa: DTZ007
                    break
                except ValueError:
                    pass
        if duration.get("endTimestamp", None) is None:
            output["end_time"] = None
        else:
            for datetime_format in datetime_formats:
                try:
                    output["end_time"] = datetime.strptime(duration.get("endTimestamp"), datetime_format)  # noqa: DTZ007
                    break
                except ValueError:
                    pass
        output["center_lat"] = dct.get("centerLatE7", None)
        output["center_lng"] = dct.get("centerLngE7", None)
        output["place_confidence"] = dct.get("placeConfidence", None)
        output["place_visit_type"] = dct.get("placeVisitType", None)
        output["visit_confidence"] = dct.get("visitConfidence", None)
        output["edit_confirmation_status"] = dct.get("editConfirmationStatus", None)
        output["place_visit_importance"] = dct.get("placeVisitImportance", None)

        parsed_locations = []
        candidate_locations = dct.get("otherCandidateLocations", [])
        for candidate_location in candidate_locations:
            loc_parsed = self._candidate_location_parser(candidate_location)
            parsed_locations.append(loc_parsed)
        output["candidate_locations"] = parsed_locations

        return output



#!/usr/bin/env python
import argparse

from dlt import pipeline as dlt_pipeline
from dlt.sources.credentials import ConnectionStringCredentials
from dlt.sources.sql_database import sql_database


def ctu_relational_to_warehouse(destination: str) -> None:
    """Run the CTU Relational database to Postgres pipeline."""
    databases = ["financial", "imdb_full", "NBA"]

    for db in databases:
        pipeline = dlt_pipeline(
            pipeline_name=f"ctu_relational_{db}_postgres",
            dataset_name=db,
            destination=destination,
        )

        credentials = ConnectionStringCredentials(
            f"mariadb+mariadbconnector://guest:ctu-relational@relational.fel.cvut.cz:3306/{db}"
        )
        source = sql_database(credentials)

        info = pipeline.run(source)
        print(info)


def main() -> None:  # noqa: D103
    parser = argparse.ArgumentParser(
        description="Transfers data from the CTU relational database to the Data Warehouse"
    )
    parser.add_argument(
        "--destination", type=str, required=True, choices=["postgres", "bigquery"], help="where the data will be saved"
    )

    args = parser.parse_args()
    ctu_relational_to_warehouse(args.destination)


if __name__ == "__main__":
    main()
